import json
import logging
import threading
import time
import uuid
from enum import Enum
from typing import Callable
from typing import Union

import pika
from flask import current_app
from flask import Flask
from retry import retry

from .encoder import UUIDEncoder

__version__ = "1.1.2"


class ExchangeTypeError(Exception):
    pass


class SyncTimeoutError(Exception):
    pass


class ExchangeType(Enum):
    """Defines all possible exchange types
    """

    DIRECT = "direct"
    """direct exchange"""

    FANOUT = "fanout"
    """fanout exchange"""

    TOPIC = "topic"
    """topic exchange"""

    HEADERS = "headers"
    """headers exchange"""


def get_state(app):
    """Gets the state for the application"""
    assert "coney" in app.extensions, (
        "The coney extension was not registered to the current "
        "application. Please make sure to call init_app() first."
    )
    return app.extensions["coney"]


class _ConeyState:
    """Remembers configuration for the (coney, app) tuple."""

    def __init__(self, coney):
        self.coney = coney
        self.connections = {}
        self.channels = {}
        self.consumer_tags = []
        self.data = {}


class Coney:
    """
    This class is used to control the Coney integration to one or more Flask
    applications. Depending on how you initialize the object it is usable right
    away or will attach as needed to a Flask application.

    There are two usage modes which work very similarly. One is binding the
    instance to a very specific Flask application::

        app = Flask(__name__)
        coney = Coney(app)

    The second possibility is to create the object once and configure the
    application later to support it::

        coney = Coney()

        def create_app():
            app = Flask(__name__)
            coney.init_app(app)
            return app

    To listen on a queue use::

        coney = Coney(app)

        @coney.queue(queue_name="test")
        def queue_test(ch, method, props, body):
            pass

    To publish a message use::

        coney = Coney(app)

        coney.publish({"test": 1})

    :param app: A flask app
    :param testing: Setup testing mode. This will not invoke threads
    """

    def __init__(self, app: Flask = None, testing: bool = False):
        self.app = app
        self.thread = None
        self.testing = testing
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        """
        This callback can be used to initialize an application for the use
        with Coney.
        """
        # We intentionally don't set self.app = app, to support multiple
        # applications. If the app is passed in the constructor,
        # we set it and don't support multiple applications.

        # noqa: B950 CONEY_URI: see https://pika.readthedocs.io/en/stable/modules/parameters.html#pika.connection.URLParameters
        if not (app.config.get("CONEY_BROKER_URI")):
            raise RuntimeError("CONEY_BROKER_URI needs to be set")

        app.extensions["coney"] = _ConeyState(self)

    def get_app(self, reference_app: Flask = None):
        """
        Helper method that implements the logic to look up an application.

        :param reference_app: A flask app
        """
        if reference_app is not None:
            return reference_app

        if current_app:
            return current_app._get_current_object()

        if self.app is not None:
            return self.app

        raise RuntimeError(
            "No application found. Either work inside a view function or push"
            " an application context. See"
            " http://mikebarkmin.github.io/flask-coney/contexts/."
        )

    @retry(pika.exceptions.AMQPConnectionError, tries=4, delay=1, jitter=3)
    def get_connection(
        self, app: Flask = None, bind: str = "__default__"
    ) -> pika.BlockingConnection:
        """Returns a specific connection. If there is no connection to the broker
        a new connection will be established.

        :param app: A flask app
        :param bind: Namespace for multiple connections
        """
        app = self.get_app(app)
        state = get_state(app)

        connection = state.connections.get(bind)
        if connection is None and app:
            params = pika.URLParameters(app.config["CONEY_BROKER_URI"])
            connection = pika.BlockingConnection(params)
            state.connections[bind] = connection

        return connection

    def get_channel(
        self, app: Flask = None, bind: str = "__default__"
    ) -> pika.channel.Channel:
        """Returns a specific channel. If there is no connection to the broker a
        new connection will be established.

        :param app: A flask app
        :param bind: Namespace for multiple connections
        """

        app = self.get_app(app)
        state = get_state(app)

        connection = self.get_connection(app=app, bind=bind)

        channel = state.channels.get(bind)
        if channel is None:
            channel = connection.channel()
            state.channels[bind] = channel

        return channel

    def close(self, app: Flask = None, bind: str = "__default__"):
        """Closes the connection

        :param app: A flask app
        :param bind: Namespace for multiple connections
        """

        app = self.get_app(app)

        connection = self.get_connection(app=app, bind=bind)
        if connection is not None:
            connection.close()

    def queue(
        self,
        queue_name: str = "",
        exchange_name: str = "",
        exchange_type: ExchangeType = ExchangeType.DIRECT,
        routing_key: str = None,
        app: Flask = None,
    ) -> Callable:
        """
        A decorator for consuming a queue. A thread will start in the
        background, if no other thread for this purpose was already started.
        There will only be one thread for every queue.

        Example::

            @coney.queue(queue_name="test")
            def queue_test(ch, method, props, body):
                pass

        :param type: ExchangeType
        :param queue_name: Name of the queue
        :param exchange_name: Name of the exchange
        :param exchange_type: Type of the exchange
        :param routing_key: The routing key
        :param app: A flask app
        :param bind: Namespace for multiple connections
        """
        app = self.get_app(app)
        state = get_state(app)

        if (
            exchange_type == ExchangeType.FANOUT
            or exchange_type == ExchangeType.DIRECT
            or exchange_type == ExchangeType.TOPIC
            or exchange_type == ExchangeType.HEADERS
        ):
            if not queue_name:
                # If queue name is empty, then declare a temporary queue
                queue_name = self._temporary_queue_declare(app=app)
            else:
                self._queue_declare(queue_name=queue_name)
        else:
            raise ExchangeTypeError(f"Exchange type {exchange_type} is not supported")

        # Consume the queue
        self._exchange_bind_to_queue(
            exchange_type=exchange_type,
            exchange_name=exchange_name,
            routing_key=routing_key,
            queue=queue_name,
            app=app,
        )

        def decorator(func):
            consumer_tag = self._basic_consuming(queue_name, func)
            state.consumer_tags.append(consumer_tag)
            return func

        # start thread if not already started, which runs in the
        # background. The background thread is only required for
        # queue, therefore it is started here.
        self.run()

        return decorator

    def _temporary_queue_declare(self, app: Flask = None):
        return self._queue_declare(exclusive=True, auto_delete=True, app=app)

    def _queue_declare(
        self,
        queue_name: str = "",
        passive: bool = False,
        durable: bool = False,
        exclusive: bool = False,
        auto_delete: bool = False,
        arguments: dict = None,
        app: Flask = None,
    ):
        channel = self.get_channel(app)
        result = channel.queue_declare(
            queue=queue_name,
            passive=passive,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            arguments=arguments,
        )
        return result.method.queue

    def _exchange_bind_to_queue(
        self,
        exchange_name: str = "",
        exchange_type: ExchangeType = ExchangeType.DIRECT,
        routing_key: str = None,
        queue: str = "",
        app: Flask = None,
    ):
        """
        Declare exchange and bind queue to exchange
        :param type: The type of exchange
        :param exchange_name: The name of exchange
        :param exchange_type: The type of exchange
        :param routing_key: The key of exchange bind to queue
        :param queue: queue name
        """
        if exchange_name == "" and routing_key is not None and routing_key != queue:
            raise RuntimeError(
                """The routing key of a queue in the default
                exchange needs to be the same as the queue name or
                None"""
            )

        if routing_key is None:
            routing_key = queue

        channel = self.get_channel(app)
        if exchange_name != "":
            channel.exchange_declare(
                exchange=exchange_name, exchange_type=exchange_type.value
            )
            channel.queue_bind(
                queue=queue, exchange=exchange_name, routing_key=routing_key
            )

    def _accept(self, corr_id: str, result: str, app: Flask = None):
        app = self.get_app(app)
        data = get_state(app).data
        data[corr_id]["is_accept"] = True
        data[corr_id]["result"] = result
        self.get_channel(app).queue_delete(data[corr_id]["reply_queue_name"])

    def _on_response(
        self,
        ch: pika.channel.Channel,
        method: pika.spec.Basic.Deliver,
        props: pika.spec.BasicProperties,
        body: str,
        app=None,
    ):
        logging.info(f"on response => {body}")

        corr_id = props.correlation_id
        if props.content_type == "application/json":
            body = json.loads(body)

        self._accept(corr_id, body, app=app)

    def _basic_consuming(
        self,
        queue_name: str,
        callback: Callable[
            [
                pika.channel.Channel,
                pika.spec.Basic.Deliver,
                pika.spec.BasicProperties,
                str,
            ],
            None,
        ],
        app=None,
    ):
        """
        Consume messages of a queue

        :param queue_name: Name of the queue
        :param callback: Function to call on new messages

        :return: Consumer tag which may be used to cancel the consumer
        """

        def advanced_callback(ch, method, props, body):
            if props.content_type == "application/json":
                body = json.loads(body)
            callback(ch, method, props, body)

        self.get_channel(app).basic_qos(prefetch_count=1)
        return self.get_channel(app).basic_consume(queue_name, advanced_callback)

    def _start_consuming(self, app=None):
        """
        Processes I/O events and dispatches timers and basic_consume
        callbacks until all consumers are cancelled.
        """
        self.get_channel(app, bind="__thread__").start_consuming()

    def _stop_consuming(self, app=None):
        self.get_channel(app, bind="__thread__").stop_consuming()

    def publish(
        self,
        body: Union[str, dict],
        exchange_name: str = "",
        routing_key: str = None,
        durable: bool = False,
        properties: dict = None,
        app: Flask = None,
    ):
        """
        Will publish a message

        Example::

            @app.route('/process'):
            def process():
                coney.publish({"text": "process me"})

        :param body: Body of the message, either a string or a dict
        :param exchange_name: The exchange
        :param exchange_type: The type of the exchange
        :param routing_key: The routing key
        :param durable: Should the exchange be durable
        :param app: A flask app
        """
        channel = self.get_channel(app)

        if properties is None:
            properties = {}

        if isinstance(body, dict):
            body = json.dumps(body, cls=UUIDEncoder)
            properties["content_type"] = "application/json"

        channel.basic_publish(
            exchange=exchange_name,
            routing_key=routing_key,
            body=body,
            properties=pika.BasicProperties(**properties),
        )

    def reply_sync(
        self,
        ch: pika.channel.Channel,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: str,
        app=None,
    ):
        """
        Will reply to a message, which was send by :meth:`publish_sync`

        Example::

            @queue(queue_name="rpc")
            def concat_callback(ch, method, props, body):
                result = body["a"] + body["b"]
                body = {"result": result}
                coney.reply_sync(ch, method, props, body)

        This is a conveniences short hand method for::

            @queue(queue_name="rpc")
            def concat_callback(ch, method, props, body):
                result = body["a"] + body["b"]
                body = {"result": result}
                ch.basic_ack(delivery_tag=method.delivery_tag)
                self.publish(
                    body,
                    routing_key=properties.reply_to,
                    properties={"correlation_id": properties.correlation_id},
                    app=app,
                )

        :parameter ch:
        :parameter method:
        :parameter properties:
        :parameter body: The message to send

        """
        ch.basic_ack(delivery_tag=method.delivery_tag)
        self.publish(
            body,
            routing_key=properties.reply_to,
            properties={"correlation_id": properties.correlation_id},
            app=app,
        )

    def publish_sync(
        self,
        body: Union[str, dict],
        exchange_name: str = "",
        routing_key: str = None,
        properties: dict = None,
        timeout: float = 10,
        app: Flask = None,
    ):
        """
        Will publish a message and wait for the response

        Example::

            # client
            @app.route('/concat')
            def concat():
                a = request.args.get('a')
                b = request.args.get('b')

                body = {'a': a, 'b': b}
                result = coney.publish_sync(body, routing_key="rpc")
                return result

            # server
            @queue(queue_name="rpc")
            def concat_callback(ch, method, props, body):
                result = body["a"] + body["b"]
                body = {"result": result}
                coney.reply_sync(ch, method, props, body)

        :param body: Body of the message, either a string or a dict
        :param exchange_name: The exchange
        :param routing_key: The routing key
        :param properties: see :py:class:`pika.spec.BasicProperties`
        :param timeout: Timeout in seconds
        :param app: A flask app
        :raises:
            SyncTimeoutError: if no message received in timeout
        """
        app = self.get_app(app)
        corr_id = str(uuid.uuid4())
        callback_queue = self._temporary_queue_declare(app=app)
        state = get_state(app)
        state.data[corr_id] = {
            "is_accept": False,
            "result": None,
            "reply_queue_name": callback_queue,
        }

        if properties is None:
            properties = {}

        if isinstance(body, dict):
            body = json.dumps(body, cls=UUIDEncoder)
            properties["content_type"] = "application/json"

        channel = self.get_channel(app)
        channel.basic_consume(callback_queue, self._on_response, auto_ack=True)
        channel.basic_publish(
            exchange="",
            routing_key=routing_key,
            body=body,
            properties=pika.BasicProperties(
                **properties, reply_to=callback_queue, correlation_id=corr_id
            ),
        )

        end = time.time() + timeout

        while time.time() < end:
            if state.data[corr_id]["is_accept"]:
                logging.info("Got the RPC server response")
                return state.data[corr_id]["result"]
            else:
                connection = self.get_connection(app)
                connection.process_data_events()
                time.sleep(0.3)
        raise SyncTimeoutError()

    def run(self):
        logging.info(" * The Flask Coney application is consuming")
        if not self.testing and (self.thread is None or not self.thread.is_alive()):
            self.thread = threading.Thread(target=self._start_consuming)
            self.thread.start()



from pallets_sphinx_themes import get_version
from pallets_sphinx_themes import ProjectLink

# Project ---------------------------------------------------------------------

project = "Flask Coney"
copyright = "2020, Mike Barkmin"
author = "Mike Barkmin"
release, version = get_version("Flask-Coney", version_length=1)

# General ---------------------------------------------------------------------

master_doc = "index"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "pallets_sphinx_themes",
    "sphinx_autodoc_typehints",
    "sphinx_issues",
]


intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "flask": ("https://flask.palletsprojects.com/en/1.1.x/", None),
    "pika": ("https://pika.readthedocs.io/en/stable/", None),
}

issues_github_path = "mikebarkmin/flask-coney"


# -- Options for HTML output ----------------------------------------------

html_theme = "flask"
html_context = {
    "project_links": [
        ProjectLink("Website", "https://www.barkmin.eu"),
        ProjectLink("PyPI releases", "https://pypi.org/project/Flask-Coney/"),
        ProjectLink("Source Code", "https://github.com/mikebarkmin/flask-coney/"),
        ProjectLink(
            "Issue Tracker", "https://github.com/mikebarkmin/flask-coney/issues/"
        ),
    ]
}
html_sidebars = {
    "index": ["project.html", "localtoc.html", "searchbox.html"],
    "**": ["localtoc.html", "relations.html", "searchbox.html"],
}
singlehtml_sidebars = {"index": ["project.html", "localtoc.html"]}
html_favicon = "_static/flask-coney-logo.png"
html_logo = "_static/flask-coney-logo.png"
html_title = f"Flask-Coney Documentation ({version})"
html_static_path = ["_static"]
html_show_sourcelink = False

# LaTeX ----------------------------------------------------------------

latex_documents = [
    (master_doc, f"Flask-Coney-{version}.tex", html_title, author, "manual",)
]



import pytest
from flask import Flask
from pytest_rabbitmq import factories

from flask_coney import Coney

rabbitmq_proc = factories.rabbitmq_proc()
rabbitmq = factories.rabbitmq("rabbitmq_proc")


@pytest.fixture
def app():
    app = Flask(__name__)
    app.testing = True
    app.config["CONEY_BROKER_URI"] = "amqp://guest:guest@localhost:6789"
    return app


@pytest.fixture
def coney(app):
    with app.app_context():
        yield Coney(app, testing=True)


@pytest.fixture
def client(app):
    with app.test_client() as client:
        yield client



import json
import time

import pika
import pytest
from flask import Flask
from rabbitpy import Exchange
from rabbitpy import Message
from rabbitpy import Queue

from flask_coney import Coney
from flask_coney import ExchangeTypeError
from flask_coney import SyncTimeoutError


def test_coney_broker_uri_not_set():
    app = Flask(__name__)
    with pytest.raises(RuntimeError):
        Coney(app)


def test_get_app_current_app(app, coney):
    my_app = coney.get_app()

    assert app == my_app


def test_get_app_property(app):
    coney = Coney(app)

    assert coney.app == app

    my_app = coney.get_app()

    assert my_app == app


def test_get_app_reference(app, coney):
    my_app = coney.get_app(reference_app=app)

    assert app == my_app


def test_get_app_missing():
    coney = Coney()

    with pytest.raises(RuntimeError):
        coney.get_app()


def test_get_connection(coney):
    with pytest.raises(pika.exceptions.AMQPConnectionError):
        coney.get_connection()


def test_close_connection(app, coney, rabbitmq, rabbitmq_proc):
    state = app.extensions["coney"]

    assert state.connections.get("default") is None

    coney.publish("Hallo", routing_key="test")

    connection = state.connections.get("__default__")
    assert connection is not None
    assert connection.is_open

    coney.close()

    assert connection.is_closed


def test_init_app(app):
    coney = Coney()
    coney.init_app(app)

    assert coney.app is None
    assert app.extensions["coney"] is not None


def test_queue_temporary(coney, rabbitmq_proc):
    @coney.queue()
    def tmp():
        pass

    gen = [q for q in rabbitmq_proc.list_queues() if "amq.gen" in q]

    assert len(gen) == 1


def test_queue_default_exchange(rabbitmq, rabbitmq_proc, app, coney):
    @coney.queue(queue_name="test")
    def test_queue(ch, method, props, body):
        pass

    queues = rabbitmq_proc.list_queues()
    assert "test" in queues


def test_queue_default_exchange_routing_key_mismatch(coney):
    with pytest.raises(RuntimeError):

        @coney.queue(queue_name="test", routing_key="test2")
        def test_queue(ch, method, props, body):
            pass


def test_queue_custom_exchange(rabbitmq, rabbitmq_proc, coney):
    @coney.queue(queue_name="hi", exchange_name="hu")
    def hi_queue(ch, method, props, body):
        assert b"dadada" == body
        coney._stop_consuming()

    exchanges = rabbitmq_proc.list_exchanges()
    assert "hu" in exchanges

    queues = rabbitmq_proc.list_queues()
    assert "hi" in queues

    channel = rabbitmq.channel()
    message = Message(channel, "dadada")
    message.publish("hu", "hi")

    coney._start_consuming()


def test_queue_custom_routing_key(rabbitmq, rabbitmq_proc, coney):
    @coney.queue(queue_name="hi", exchange_name="hu", routing_key="custom")
    def hi_queue(ch, method, props, body):
        assert b"dadada" == body
        coney._stop_consuming()

    exchanges = rabbitmq_proc.list_exchanges()
    assert "hu" in exchanges

    queues = rabbitmq_proc.list_queues()
    assert "hi" in queues

    channel = rabbitmq.channel()
    message = Message(channel, "dadada")
    message.publish("hu", "custom")

    coney._start_consuming()


def test_queue_json(rabbitmq, coney):
    @coney.queue(queue_name="hi", exchange_name="hu", routing_key="custom")
    def hi_queue(ch, method, props, body):
        response = {"Hi": "hu"}
        assert response == body
        coney._stop_consuming()

    channel = rabbitmq.channel()
    message = Message(
        channel,
        json.dumps({"Hi": "hu"}),
        properties={"content_type": "application/json"},
    )
    message.publish("hu", "custom")
    coney._start_consuming()


def test_queue_wrong_exchange_type(coney, rabbitmq, rabbitmq_proc):
    with pytest.raises(ExchangeTypeError):

        @coney.queue(exchange_name="hallo", exchange_type="not-right")
        def hallo_queue(ch, method, props, body):
            pass


def test_publish(rabbitmq, coney):
    channel = rabbitmq.channel()
    exchange = Exchange(channel, "my-exchange", "direct")
    exchange.declare()
    queue = Queue(channel, "my-queue")
    queue.declare()
    queue.bind(exchange, "my-routing-key")

    coney.publish("Hi", exchange_name="my-exchange", routing_key="my-routing-key")

    for message in queue.consume():
        assert message.body == b"Hi"
        queue.stop_consuming()


def test_publish_sync(coney):
    @coney.queue(queue_name="echo", exchange_name="sync")
    def echo_queue(ch, method, props, body):
        body = {**body, "echo": True}
        coney.reply_sync(ch, method, props, body)

    coney.testing = False
    coney.run()

    time.sleep(1)

    result = coney.publish_sync({"Hi": "Ho"}, exchange_name="sync", routing_key="echo")

    assert result == {"Hi": "Ho", "echo": True}


def test_publish_sync_timeout(coney):

    with pytest.raises(SyncTimeoutError):
        coney.publish_sync("Hi", routing_key="test", timeout=0.5)



from flask import Flask, request, jsonify
from flask_coney import Coney

app = Flask(__name__)
app.config["CONEY_BROKER_URI"] = "amqp://user:password@rabbitmq"
coney = Coney(app)

processed_data = []


@app.route("/processed", methods=["GET"])
def processed():
    return jsonify(processed_data)


@coney.queue(queue_name="process")
def process_queue(ch, method, params, body):
    # do something with body
    processed_body = body
    processed_data.append(processed)




from uuid import uuid4

import pytest

from flask_coney.encoder import UUIDEncoder


def test_encode_uuid():
    id = uuid4()

    encoded = UUIDEncoder().default(id)

    assert str(id) in encoded


def test_encode_not_jsonable():
    encode = [1, 2, 3, 4]

    with pytest.raises(TypeError):
        UUIDEncoder().default(encode)



from flask import Flask, request
from flask_coney import Coney

app = Flask(__name__)
app.config["CONEY_BROKER_URI"] = "amqp://user:password@rabbitmq"
coney = Coney(app)


@app.route("/process", methods=["POST"])
def process():
    data = request.get_json()
    # validation ...
    coney.publish(data, routing_key="process")




import re

from setuptools import setup

with open("src/flask_coney/__init__.py", encoding="utf8") as f:
    match = re.search(r'__version__ = "(.*?)"', f.read())
    if not match:
        raise RuntimeError("Version could not be found")
    version = match.group(1)

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="Flask-Coney",
    version=version,
    install_requires=["Flask>=1.0.4", "Pika>=1.1.0", "retry>=0.9.2"],
)



import json
from uuid import UUID


class UUIDEncoder(json.JSONEncoder):
    """Custom JSONEncoder to support UUID objects"""

    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return str(obj)
        return json.JSONEncoder.default(self, obj)
